//*****************************************************
// Observer-Based Control
// Aoyama Satoru                 Date:2010.04.11    @
//*****************************************************

#include <p30f2012.h>
#include <adc12.h>
#include <timer.h>
#include <outcompare.h>
#include <dsp.h>
#include <uart.h> 
  
// configuration
_FWDT(WDT_OFF); 
_FGS(CODE_PROT_OFF); 
_FOSC(CSW_FSCM_OFF & FRC_PLL16);
_FBORPOR(PBOR_OFF & PWRT_64 & MCLR_EN);

// UART control register
unsigned int _U1BRG;
unsigned int _U1MODE; 
unsigned int _U1STA;
#define Fcy 7370000*4
#define BaudRate 115200
 
// A/D control register
unsigned int _ADCON1;
unsigned int _ADCON2;
unsigned int _ADCON3;
unsigned int _ADPCFG;
unsigned int _ADCSSL;

// PWM control register
unsigned int _PTCON;
unsigned int _PWMCON1;
unsigned int _PWMCON2;

// data for UART
int ResultData;
char c0, c1, c2, c3, c4, c5;
char d0, d1, d2, d3, d4, d5;

// data
int   	volume_rotary=0;
int   	gyroscope=0;
int   	sensor_reserve=0;
float   ControlOutput=0.0f;
int   	Output=0;
float 	center=0.0f;
float 	centerb=0.0f;
float 	centerE=0;
float   Error=0.0f;
float   Errorb=0.0f;
float   ErrorE=0.0f;
float 	posi=0.0f;
float 	trolley_v=0.0f;
float 	angular_v=0.0f;
float 	angle=0.0f;
int 	ref=1500; // 1500.0f = 3276 * 0.46[m]
float 	outdatad=0.0f; // for volume_rotary 
int   	flag=0;

//int DelayLineA[5]= {0,0,0,0,0}; //for sensor_reserve
int DelayLineb[5]= {0,0,0,0,0}; // for volume_rotary
int DelayLineE[5]= {0,0,0,0,0}; // for gyroscope
int DelayLined[2]= {0,0}; // for volume_rotary
int Coeffc[2]= {32767,-32767};  
int Coeff[5] = {6553,6553,6553,6553,6553}; 

int MatrixA[16]=  {	32767,	     1,	  321,	    0,
					    0,	 32681,	    0,	  327,
					    0,	   252,	31483,	    1,
					    0,	-17308,	    0,	32616  };

int Matrixbh[8]=  { 	5,    4731,
				      -27,   -3364,
				     1071,   -4060,
				    -5451,   15627  }; 

int MatrixZ[4]= {0,0,0,0}; 
int MatrixU[2]= {0,0}; 
int Matrixtmpa[4]= {0,0,0,0}; 
int Matrixtmpb[4]= {0,0,0,0}; 

void Push5Elems(int din, int *line){
	int tmpDelayLine[5];
	VectorCopy(5-1, tmpDelayLine+1, line);
	tmpDelayLine[0] = din;
	VectorCopy(5, line, tmpDelayLine);
}

void Push2Elems(int din, int *line){
	int tmpDelayLine[2];
	VectorCopy(2-1, tmpDelayLine+1, line);
	tmpDelayLine[0] = din;
	VectorCopy(2, line, tmpDelayLine);
}

// Timer subroutine
void __attribute__((__interrupt__,__shadow__, no_auto_psv)) _AltT3Interrupt(void)
{
	IFS0bits.ADIF=0;
	sensor_reserve = ReadADC12(0); // not in use. for reserve AN0 PinNo.8
	volume_rotary = ReadADC12(1); // volume_rotary   AN1 PinNo.9
	gyroscope    = ReadADC12(2); // gyroscope        AN2 PinNo.10

	if(flag <50 ){
		flag = flag + 1;	
		//Push5Elems(sensor_reserve, DelayLineA);
		//center  = (float)VectorDotProduct(5, DelayLineA, Coeff);
     
		Push5Elems(volume_rotary, DelayLineb);
		centerb = (float)VectorDotProduct(5, DelayLineb, Coeff);

		Push5Elems(gyroscope, DelayLineE);
		centerE= (float)VectorDotProduct(5, DelayLineE, Coeff);
		}
	if(flag >49){	

		//trolley	
		Push5Elems(volume_rotary, DelayLineb);
		Errorb =(float)VectorDotProduct(5, DelayLineb, Coeff) -  (float)centerb;
		Push2Elems(Errorb, DelayLined);
		outdatad =(float)VectorDotProduct(2, DelayLined, Coeffc); 
 		posi = 6.28*0.030*6.0*(float)Errorb / 4096.0; // 2.0*3.14*0.03*6*(float)Errorb / 4096 [m]
    	trolley_v = 100.0*6.28*0.03*6.0*(float)outdatad / 4096.0; // outdatad*100*6.28*0.03*6/4096

		//gyroscope
		Push5Elems(gyroscope, DelayLineE);
		ErrorE =(float) VectorDotProduct(5, DelayLineE, Coeff) - (float)centerE;
		angular_v = 0.017456 * 4.8 * (float)ErrorE / (4096.0*0.00067); //(4.8/4096)*(0.017453/0.00067)
		angle = angle + 0.01 * (float)angular_v;

		MatrixU[0] = ControlOutput;
		MatrixU[1] =  3276.0*(float)posi      - 1.0*(float)MatrixZ[0]
					+ 3276.0*(float)angular_v - 1.0*(float)MatrixZ[3] ;

		MatrixMultiply(4,4,1,Matrixtmpa,MatrixA,MatrixZ);
		MatrixMultiply(4,2,1,Matrixtmpb,Matrixbh,MatrixU);
		MatrixAdd(4,1,MatrixZ,Matrixtmpa,Matrixtmpb);	

		ControlOutput = - 2.0*(float)(MatrixZ[0] - ref) + 4.0*(float)MatrixZ[1]
						- 0.0*(float)MatrixZ[2]         + 0.5*(float)MatrixZ[3];

		//ControlOutput = - 2.0*(float)(MatrixZ[0] - ref) ;
	     
	/*	
		if(flag <1000){
			ControlOutput = 1638; //  0.50(N)  
			flag = flag + 1;
			}
			else{
				ControlOutput =0;
			}
	*/	

	} //if(flag >49){	

		if(ControlOutput < 0){
		PORTBbits.RB7 =0;//       Pin No.31 --->  TA7291P Pin No.5
		PORTBbits.RB8 =1;//       Pin No.30 --->  TA7291P Pin No.6
		Output = -0.5*(float)ControlOutput;         
		}else{
		PORTBbits.RB7 =1;
		PORTBbits.RB8 =0;  
		Output = 0.5*(float)ControlOutput;
		}  		
		SetDCOC2PWM(Output);		// set PWM data OC2 Pin No.29--->  TA7291P Pin No.4

	if(flag >49){
	  if(flag <570){
		flag = flag + 1;

		ResultData = MatrixZ[0]; //trolley_posi
		//ResultData = MatrixZ[1]; //angle
		//ResultData = MatrixZ[2]; //trolley_v
		//ResultData = MatrixZ[3]; //angular_v
		//ResultData = MatrixU[0];
		//ResultData = MatrixU[1];

 		if(ResultData < 0){
			ResultData = -ResultData;
		c0 = 0x2d;
		}else{c0 = 0x20;}
		c1 = 0x30 + ResultData % 10;
		ResultData = ResultData / 10;		 
		c2 = 0x30 + ResultData % 10;			
		ResultData = ResultData / 10;		 
		c3 = 0x30 + ResultData % 10;		
		ResultData = ResultData / 10;		 
		c4 = 0x30 + ResultData % 10;
		ResultData = ResultData / 10;		 
		c5 = 0x30 + ResultData % 10;		

		//ResultData = 3276.0*(float)posi;
		//ResultData = 3276.0*(float)angle;
		//ResultData = 3276.0*(float)angular_v;
		//ResultData = 3276.0*(float)trolley_v;
		ResultData = MatrixZ[1]; //angle

		if(ResultData < 0){
			ResultData = -ResultData;
		d0 = 0x2d;
		}else{d0 = 0x20;}
		d1 = 0x30 + ResultData % 10;
		ResultData = ResultData / 10;		 
		d2 = 0x30 + ResultData % 10;			
		ResultData = ResultData / 10;		 
		d3 = 0x30 + ResultData % 10;
		ResultData = ResultData / 10;		 
		d4 = 0x30 + ResultData % 10;
		ResultData = ResultData / 10;		 
		d5 = 0x30 + ResultData % 10;

		while(BusyUART1());
	    WriteUART1(c0);	
	    while(BusyUART1());
	    WriteUART1(c5);
		while(BusyUART1());
		WriteUART1(c4);		
		while(BusyUART1());
		WriteUART1(c3);	
		while(BusyUART1());
		WriteUART1(c2);
	    while(BusyUART1());
	    WriteUART1(c1);

		while(BusyUART1());
		WriteUART1(0x3b);

	    while(BusyUART1());
	    WriteUART1(d0);
	    while(BusyUART1());
	    WriteUART1(d5);
		while(BusyUART1());
		WriteUART1(d4);	
		while(BusyUART1());
		WriteUART1(d3);	
		while(BusyUART1());
		WriteUART1(d2);
	    while(BusyUART1());
	    WriteUART1(d1);

		while(BusyUART1());
		WriteUART1(0x0d);
	    while(BusyUART1());
	    WriteUART1(0x0a);
	  }
	}

	IFS0bits.T3IF = 0;			// clear T3 interuppt flag
}

// Main routine
int main(void)
{
    int i,j;
	for(i=0; i<400; i=i+1){
		 for(j=0; j<6500; j=j+1){ ; }
 	}  

	// Port Output:  PWM RB9 ,  motor revers RB7 & RB8
	TRISB=0x007F;
	//Use Alternate Interrupt Vector Table
	INTCON2bits.ALTIVT=1;

	CloseUART1();
	ConfigIntUART1(	UART_RX_INT_DIS & UART_TX_INT_DIS);
	// BaudRate=Fcy/(16(U1BRG+1))
	//         =29.48MHz/(16(15+1))
	//         =115156
	// Error=(115156-115200)/115200
	//      =-0.0382%
	_U1BRG=	((Fcy/BaudRate)/16);
	// alternate I/O (U1ATX/U1ARX), N81
	_U1MODE=	UART_EN &
			UART_IDLE_CON &
			UART_ALTRX_ALTTX &
			UART_DIS_WAKE &
			UART_DIS_LOOPBACK &
			UART_EN_ABAUD &
			UART_NO_PAR_8BIT &
			UART_1STOPBIT;
	_U1STA=	UART_INT_TX_BUF_EMPTY &
			UART_TX_PIN_NORMAL &
			UART_TX_ENABLE &
			UART_INT_RX_CHAR &
			UART_ADR_DETECT_DIS &
			UART_RX_OVERRUN_CLEAR;
	OpenUART1(_U1MODE, _U1STA, _U1BRG);

	// Confirm to turn off ADC
	ADCON1bits.ADON=0; // turn off ADC
	// ADC init	
	_ADCON1=ADC_MODULE_ON & 
 			ADC_IDLE_CONTINUE & 
			ADC_FORMAT_INTG & 
			ADC_CLK_TMR &
			//ADC_CLK_AUTO &
			ADC_AUTO_SAMPLING_ON & 
			ADC_SAMP_OFF;
	_ADCON2=ADC_VREF_AVDD_AVSS & 
			//ADC_SCAN_OFF &
			ADC_SCAN_ON &
			ADC_SAMPLES_PER_INT_3 &
			ADC_ALT_BUF_OFF & 
			ADC_ALT_INPUT_OFF;
	_ADCON3=ADC_SAMPLE_TIME_14 & 
			ADC_CONV_CLK_SYSTEM & 
			ADC_CONV_CLK_32Tcy;
	_ADPCFG=ENABLE_AN0_ANA & 
			ENABLE_AN1_ANA &  
		    ENABLE_AN2_ANA;	
	_ADCSSL=0x007;				// A-D channel-3 analog input ---> ADC_SAMPLES_PER_INT_3 
	OpenADC12(_ADCON1, _ADCON2, _ADCON3, _ADPCFG, _ADCSSL);

	// Timer3 setting sampling ADC
	OpenTimer3(T3_ON & T3_GATE_OFF & T3_PS_1_64 & T3_SOURCE_INT, 4606);	// sampling 0.01 sec
	ConfigIntTimer3(T3_INT_PRIOR_5 & T3_INT_ON);
	// Timer2 setting carrier PWM
	OpenTimer2(T2_ON & T2_GATE_OFF & T2_PS_1_1 & T2_SOURCE_INT, 650);	// PWM 
	OpenOC2(OC_IDLE_CON & OC_TIMER2_SRC & OC_PWM_FAULT_PIN_DISABLE, 0, 0);
	// enable interrupt
	DISICNT = 0x0000;

	// main loop
	while(1)
	{  

	}//while(1)
}
